#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>
#include <math.h>
#include <mpi.h>
#include "physical_constants.h"
#include "pbc.h"



int main(int argc, char** argv)
{

  // 64bit system .. ... .... .... . ....
  // ... .. float .. float .. ..

  FILE *info, *info2;
  /* field vars */
  float *dummy,*dummy_pml;//,*t_field;
  float *Ex,*Ey,*Ez;
  float *Hx,*Hy,*Hz;
  float *rFx,*rFy,*rFz;
  float *rFxy,*rFyx,*rFzx;
  float *rFxz,*rFyz,*rFzy;
  float *Euxy, *Euxz, *Euyx, *Euyz, *Euzx, *Euzy;
  float *Huxy, *Huxz, *Huyx, *Huyz, *Huzx, *Huzy;
  float *Ebxy, *Ebxz, *Ebyx, *Ebyz, *Ebzx, *Ebzy;
  float *Hbxy, *Hbxz, *Hbyx, *Hbyz, *Hbzx, *Hbzy;

  /* coef. vars */
  float indexs[4];
  float a,r,d,t,freq,dx,dt,inj,fnz,k1,k2;
  float *Cb,Db;
  float R1,R2;
  float *sigma,sigma_max,m=3.5;
  int i,j,k;
  int nx,ny,nz,x0,y0,z0,maxn;
  int sx,sy,sz;
  int type;
  int node,size;
  int pos=0;
  long int iter,max_iter;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &node);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  //sigma_max = (m+1)*logf(10e-6)/(2*PML*sqrt(MU_0/EPSILON_0));
  sigma_max = 142000;
  sigma = (float *)malloc(sizeof(float)*PML);
  for(i=0;i<PML;i++)
    sigma[i] = pow(((float)i+1)/(PML),m)*sigma_max;


  // Step 1
  LOAD_INFO_FILE;

  if(node ==0)
    printf("\n%f\n%f\n%f\n%f\n",indexs[0],indexs[1],indexs[2],indexs[3]);

  // Step 2
  MAKE_STRUCTURE;


  // Step 3
  //CALCULATION
  if(node == 0){
    printf("R1 = %e\tR2 = %e\n",R1,R2);
    printf("# Iteration start\n");                                
    info = fopen("pos1.dat","w");                              
    info2 = fopen("pos2.dat","w");                              
  }

  dummy = (float *)malloc(sizeof(float)*maxn*maxn*3);
  dummy_pml = (float *)malloc(sizeof(float)*maxn*PML*9);
  rFx = (float *)malloc(sizeof(float)*maxn*maxn);
  rFy = (float *)malloc(sizeof(float)*maxn*maxn);
  rFz = (float *)malloc(sizeof(float)*maxn*maxn);
  rFxy = (float *)malloc(sizeof(float)*maxn*PML*3);
  rFxz = (float *)malloc(sizeof(float)*maxn*PML*3);
  rFyx = (float *)malloc(sizeof(float)*maxn*PML*3);
  rFyz = (float *)malloc(sizeof(float)*maxn*PML*3);
  rFzx = (float *)malloc(sizeof(float)*maxn*PML*3);
  rFzy = (float *)malloc(sizeof(float)*maxn*PML*3);
  bzero(dummy,sizeof(float)*maxn*maxn*3);	        
  bzero(dummy_pml,sizeof(float)*maxn*PML*9);	        
  bzero(rFx,sizeof(float)*maxn*maxn);	            
  bzero(rFy,sizeof(float)*maxn*maxn);	           
  bzero(rFz,sizeof(float)*maxn*maxn);	          
  bzero(rFxy,sizeof(float)*maxn*PML*3);	            
  bzero(rFxz,sizeof(float)*maxn*PML*3);	            
  bzero(rFyx,sizeof(float)*maxn*PML*3);	           
  bzero(rFyz,sizeof(float)*maxn*PML*3);	           
  bzero(rFzx,sizeof(float)*maxn*PML*3);	          
  bzero(rFzy,sizeof(float)*maxn*PML*3);	          

  ////////////////////////////////////////
  // Iteration start
  //
  if(node == 0)	printf("k1 = %f\nk2 = %f\n",k1*a/2/M_PI,k2*a/2/M_PI);
  for(iter=0;iter<max_iter;iter++)
  {
    if(node==0&&iter%10==0)
    {
      printf("\rIteration : %ld/%ld",iter+10,max_iter);
      fflush(stdout);
    }
    // PML
    for(i=1;i<nx;i++)
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  if(k==0){
	    E_FIELD_PML_UP_0;
	  }
	  else{
	    E_FIELD_PML_UP;
	    E_FIELD_PML_BELOW;
	  }
	}
    // calc. space
    for(i=1;i<nx;i++)
      for(j=1;j<ny;j++)
	for(k=0;k<nz;k++){
	  if(k==0){
	    E_FIELD_UPDATE_0;
	  }
	  else{
	    E_FIELD_UPDATE;
	  }
	}
    // PBC
    if(node==0)
    {
      // bloch boundary
      j=ny-1;
      Packi;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      Unpacki;
      for(i=1;i<nx;i++)
	for(k=0;k<nz;k++){
	  E_FIELD_UPDATE_J0_1;
	}

      i=nx-1;
      Packj;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      Unpackj;
      for(j=1;j<ny;j++)
	for(k=0;k<nz;k++){
	  E_FIELD_UPDATE_I0_1;
	}

      // PML bloch boundary
      j=ny-1;
      PackUPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_J0_UPML_1;
	}

      j=ny-1;
      PackBPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_J0_BPML_1;
	}

      i=nx-1;
      PackUPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_I0_UPML_1;
	}
      i=nx-1;
      PackBPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_I0_BPML_1;
	}
    }
    else
    {
      // bloch boundary
      j=ny-1;
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      Unpacki;
      Packi;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<nz;k++){
	  E_FIELD_UPDATE_J0_2;
	}

      i=nx-1;
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      Unpackj;
      Packj;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<nz;k++){
	  E_FIELD_UPDATE_I0_2;
	}

      // PML bloch boundary
      j=ny-1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      PackUPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_J0_UPML_2;
	}
      j=ny-1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      PackBPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_J0_BPML_2;
	}

      i=nx-1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      PackUPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_I0_UPML_2;
	}
      i=nx-1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      PackBPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  E_FIELD_UPDATE_I0_BPML_2;
	}
    }


    ///////////////////////
    //  H FIELD

    // PML
    for(i=1;i<nx-1;i++)
      for(j=1;j<ny-1;j++)
	for(k=0;k<PML;k++)
	{
	  if(k==PML-1){
	    H_FIELD_PML_BELOW_L;
	  }
	  else if(k==0){
	    H_FIELD_PML_UP;
	  }
	  else{
	    H_FIELD_PML_UP;
	    H_FIELD_PML_BELOW;
	  }
	}
    // calc. space
    for(i=1;i<nx-1;i++)
      for(j=1;j<ny-1;j++)
	for(k=0;k<nz;k++)
	{
	  if(k==nz-1)
	  {
	    H_FIELD_UPDATE_L;
	  }
	  else
	  {
	    H_FIELD_UPDATE;
	  }
	}
    // PBC
    if(node==0)
    {
      // bloch boundary
      j=1;
      PackHi;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackHi;
      for(i=1;i<nx;i++)
	for(k=0;k<nz;k++){
	  H_FIELD_UPDATE_NJ_1;
	}

      i=1;
      PackHj;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackHj;
      for(j=1;j<ny;j++)
	for(k=0;k<nz;k++){
	  H_FIELD_UPDATE_NI_1;
	}

      // PML bloch boundary
      j=1;
      PackHUPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NJ_UPML_1;
	}
      j=1;
      PackHBPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NJ_BPML_1;
	}

      i=1;
      PackHUPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NI_UPML_1;
	}
      i=1;
      PackHBPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,1,0,MPI_COMM_WORLD);
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,1,1,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NI_BPML_1;
	}
    }
    else
    {
      // bloch boundary
      j=1;
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackHi;
      PackHi;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<nz;k++){
	  H_FIELD_UPDATE_NJ_2;
	}

      i=1;
      MPI_Recv(dummy,maxn*maxn*3,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackHj;
      PackHj;
      MPI_Send(dummy,maxn*maxn*3,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<nz;k++){
	  H_FIELD_UPDATE_NI_2;
	}

      // PML bloch boundary
      j=1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      PackHUPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NJ_UPML_2;
	}
      j=1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLi;
      PackHBPMLi;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(i=1;i<nx;i++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NJ_BPML_2;
	}

      i=1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      PackHUPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NI_UPML_2;
	}
      i=1;
      MPI_Recv(dummy_pml,maxn*PML*9,MPI_FLOAT,0,0,MPI_COMM_WORLD,&status);
      UnpackPMLj;
      PackHBPMLj;
      MPI_Send(dummy_pml,maxn*PML*9,MPI_FLOAT,0,1,MPI_COMM_WORLD);
      for(j=1;j<ny;j++)
	for(k=0;k<PML;k++){
	  H_FIELD_UPDATE_NI_BPML_2;
	}
    }

    /*
    if(iter == 50)
      inj = 1.0;
    else
      inj=0.0;
      */
    inj=exp(-SQUARE(freq*130*iter*dt/5.- 6))*sin(2*M_PI*freq*iter*dt);
    //inj=sin(2*M_PI*freq*iter*dt);        
    Ex[posit(sx+1,sy+1,sz)] +=inj;
    Ey[posit(sx+1,sy+1,sz)] +=inj;
    //if(node==0) fprintf(info,"%e\t%e\n",dt*iter,inj); 
    if(node==0){// && iter>=8192){
      fprintf(info,"%e\t%e\n",dt*iter,Hz[posit(nx-(sx+1),(sy+1),sz)]); 
      //t_field[iter] = Hz[posit(33,72,z0)];
      fprintf(info2,"%e\t%e\n",dt*iter,Hz[posit(x0+1,y0+2,sz)]); 
    }
  }// end of iteration
  if(node==0)
  {
    printf("\n");
    fclose(info);
    fclose(info2);
  }


  // data save
  if(node==0)
  {
    info = fopen("field.dat","w");
    for(i=1;i<nx-1;i++)
    {
      for(j=1;j<ny-1;j++)
	fprintf(info,"%e ",Hz[posit(i,j,z0)]);
      fprintf(info,"\n");
    }
    fclose(info);
    info = fopen("cut.dat","w");
    for(i=1;i<nx-1;i++)
    {
      for(k=0;k<nz;k++)
	fprintf(info,"%e ",Hz[posit(i,y0,k)]);
      fprintf(info,"\n");
    }
    fclose(info);

    info = fopen("Cb.dat","w");
    for(i=1;i<nx-1;i++)
    {
      for(j=1;j<ny-1;j++)
	fprintf(info,"%e ",Cb[posit(i,j,z0)]);
      fprintf(info,"\n");
    }
    fclose(info);
    info = fopen("Cb_cut.dat","w");
    for(i=1;i<nx-1;i++)
    {
      for(k=0;k<nz;k++)
	fprintf(info,"%e ",Cb[posit(i,y0,k)]);
      fprintf(info,"\n");
    }
    fclose(info);
  }

  free(rFx);
  free(rFy);
  free(rFz);
  free(dummy);

  MPI_Finalize();

  return 0;
}

